/*
; PACKER92 - Copyright 1998 David Kuehling
; Adaptation for TI-89/92+ - Copyright 2002, 2003, 2004, 2005 Patrick Pelissier
;
; This file is part of the PACKER92 Library.
;
; The PACKER92 Library is free software; you can redistribute it and/or modify
; it under the terms of the GNU Lesser General Public License as published by
; the Free Software Foundation; either version 2.1 of the License, or (at your
; option) any later version.
;
; The PACKER92 Library is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
; License for more details.
;
; You should have received a copy of the GNU Lesser General Public License
; along with the PACKER92 Library; see the file COPYING.LIB.  If not, write to
; the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
; MA 02111-1307, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "macro.h"

#define MAXMATCHLENGTH 1024

#define OUTPUTCYCLE 63

/***************************************************************************
*                         Bitoperationen                                   *
***************************************************************************/

unsigned char InByte, OutByte, InBit, OutBit, OrdChrTbl[256], OrdFreqTbl[256];

unsigned MaxMDist, DifferentChars;

long ChrFreq[256], InFileLen;


FILE *OutFile, *InFile;


void
WriteBit (unsigned char bit)
{
  OutByte |= (bit << OutBit++);
  if (OutBit >= 8)
    {
      fputc (OutByte, OutFile);
      OutBit = 0;
      OutByte = 0;
    }
}

char
ReadBit2 ()
{
  register unsigned char bit;

  bit = (InByte >> InBit++) & 1;
  if (InBit >= 8)
    {
      InBit = 0;
      InByte = fgetc (InFile);
    }

  return (bit);
}

void
InitRead ()
{
  InByte = fgetc (InFile);
  InBit = 0;
}

void
InitWrite ()
{
  OutBit = 0;
  OutByte = 0;
}

void
EndWrite ()
{
  if (OutBit > 0)
    fputc (OutByte, OutFile);
}




void
WriteValue1_1 (unsigned Value)
{
  for (;;)
    {
      WriteBit (Value & 1);	/* Datenbit setzen */
      Value >>= 1;
      if (!Value)
	break;
      WriteBit (1);		/* Continue-Bit setzen */
      Value--;			/* Ausgleich der Zunahme der Mgl. */
    }
  WriteBit (0);			/* Continue-Bit lschen */
}

unsigned
ReadValue1_1 ()
{
  register unsigned Value, BitCount;
  unsigned Add = 0;

  for (BitCount = 0, Value = 0;; BitCount++)
    {
      Value |= ReadBit2 () << BitCount;
      if (!ReadBit2 ())		/* Wenn Continue-Bit gelscht - Ende */
	break;
      Add += 2 << BitCount;
    }
  return (Value + Add);
}


unsigned
getBitsValue1_1 (register unsigned v)
{
  register int Bits = 0;

  v++;
  while (v != 0)
    {
      v--;
      Bits += 2;
      v >>= 1;
    }
  return (Bits);
}


void
WriteValue3_1 (unsigned Value)
{
  for (;;)
    {
      WriteBit (Value & 1);	/* Datenbits setzen */
      WriteBit (Value >> 1 & 1);
      WriteBit (Value >> 2 & 1);
      Value >>= 3;

      if (!Value)
	break;
      WriteBit (1);		/* Continue-Bit setzen */
      Value--;			/* Ausgleich der Zunahme der Mgl. */
    }
  WriteBit (0);			/* Continue-Bit lschen */

}

unsigned
ReadValue3_1 ()
{
  register unsigned Value, BitCount;
  unsigned Add = 0;

  for (BitCount = 0, Value = 0;; BitCount += 3)
    {
      Value |= ReadBit2 () << BitCount;
      Value |= ReadBit2 () << (BitCount + 1);
      Value |= ReadBit2 () << (BitCount + 2);
      if (!ReadBit2 ())		/* Wenn Continue-Bit gelscht - Ende */
	break;
      Add += 8 << BitCount;
    }
  return (Value + Add);
}


unsigned
getBitsValue3_1 (register unsigned v)
{
  register int Bits = 0;

  v++;
  while (v != 0)
    {
      v--;
      Bits += 4;
      v >>= 3;
    }
  return (Bits);
}



/*********************************************************************
*                            Puffersystem                            *
*********************************************************************/

unsigned char *RotBuffer;

unsigned RBuffPos, RBuffEnd;

long PackedBytes, UnPackedBytes, OutStart;

void
FillRotBuffer ()
{
  while ((RBuffEnd & 32767) != ((RBuffPos + MAXMATCHLENGTH) & 32767))
    {
      RotBuffer[32767 & RBuffEnd] = fgetc (InFile);
      if (feof (InFile))
	break;
      RBuffEnd++;
    }
}


unsigned
ScanRotChar (register unsigned char c, unsigned from)
{
  register unsigned I;
  unsigned ScRBPos;
  long Bytes;

  for (I = from, ScRBPos = RBuffPos - from, Bytes = PackedBytes - from;
       I < MaxMDist && Bytes >= 0; I++, ScRBPos--, Bytes--)
    if (RotBuffer[ScRBPos & 32767] == c)
      return (I);

  return (0xFFFF);
}

unsigned
ScanRotRep (unsigned *minlen, unsigned from)
{
  register unsigned Scanp1, Scanp2;
  unsigned pos, I;

  if (*minlen >= MAXMATCHLENGTH)
    return (0xFFFF);

  for (;;)
    {
      pos = ScanRotChar (RotBuffer[RBuffPos & 32767], from);
      if (pos == 0xFFFF)
	return (0xFFFF);

      for (I = 0, Scanp1 = (RBuffPos - pos) + 1, Scanp2 = RBuffPos + 1;
	   (Scanp1 & 32767) != (RBuffPos & 32767)
	   && (Scanp2 & 32767) != (RBuffEnd & 32767) && I < MAXMATCHLENGTH;
	   Scanp1++, Scanp2++, I++)
	if (RotBuffer[Scanp1 & 32767] != RotBuffer[Scanp2 & 32767])
	  break;

      if (++I > *minlen)
	{
	  *minlen = I;
	  return (pos);
	}

      from = pos + 1;
    }
}


unsigned
ScanRotRLE ()			/* liefert die Anzahl der gleichen Bytes */
{				/* an der akt. Pos. */
  register unsigned char c;
  register int I;
  unsigned ScRBPos;

  c = RotBuffer[RBuffPos & 32767];
  for (I = 0, ScRBPos = RBuffPos;
       I < MAXMATCHLENGTH && (ScRBPos & 32767) != (RBuffEnd & 32767);
       I++, ScRBPos++)
    if (RotBuffer[ScRBPos & 32767] != c)
      break;

  return (I);
}

void
getBitsBuff (unsigned *lendat)
{
  register unsigned I, Bits, RBPos;

  for (RBPos = RBuffPos, I = 0, Bits = 0;
       (RBPos & 32767) != (RBuffEnd & 32767); I++, RBPos++)
    lendat[I] = (Bits += getBitsValue3_1 (OrdChrTbl[RotBuffer[RBPos]]));
}

void
InitRotBuffer ()
{
  if ((RotBuffer = malloc (32768)) == NULL)
    {
      printf ("\n\n\tERROR: not enough memory\n\n\t\t program aborted.\n");
      exit (1);
    }

  RBuffPos = 0;
  RBuffEnd = 0;
  PackedBytes = 0;
  FillRotBuffer ();
}


/***************************************************************************
*                            Statistik                                     *
***************************************************************************/

void
getChrFreq ()			/* Hinweis: lscht Freq, DiffChrs nicht! */
{				/*          und benutzt kein fseek. */
  int Opt, ThisOpt;
  unsigned RepLen, RepPos, PackLen;

  char Display = 0;

  InitRotBuffer ();

  while ((RBuffPos & 32767) != (RBuffEnd & 32767))
    {
      if (!(Display++ & OUTPUTCYCLE))
	{
	  printf ("\r\t Scanning... %5.1f%% completed",
		  ((double) PackedBytes * 100) / InFileLen);
	  fflush (stdout);
	}
      Opt = 0;
      RepLen = 0;
      PackLen = 0;
      RepPos = 1;
      while ((RepPos = ScanRotRep (&RepLen, ++RepPos)) != 0xFFFF)
	{
	  ThisOpt = (RepLen << 3) - getBitsValue3_1 (RepPos + 128) -
	    getBitsValue1_1 (RepLen);
	  if (ThisOpt > Opt)
	    {
	      PackLen = RepLen;
	      Opt = ThisOpt;
	    }
	}

      RepLen = ScanRotRLE ();
      ThisOpt = (RepLen << 3) - getBitsValue1_1 (RepLen) - 22;
      if (ThisOpt > Opt && ThisOpt > 2)
	{
	  if (!ChrFreq[RotBuffer[RBuffPos & 32767]])
	    DifferentChars++;

	  ChrFreq[RotBuffer[RBuffPos & 32767]]++;

	  RBuffPos += RepLen;
	  PackedBytes += RepLen;
	}
      else if (Opt > 2)
	{
	  RBuffPos += PackLen;
	  PackedBytes += PackLen;
	}
      else
	{
	  if (!ChrFreq[RotBuffer[RBuffPos & 32767]])
	    DifferentChars++;

	  ChrFreq[RotBuffer[RBuffPos++ & 32767]]++;
	  PackedBytes++;
	}

      FillRotBuffer ();
    }
  free (RotBuffer);
}


void
getOrdChrFreqTbl ()
{
  register int I, I2;
  long max;
  int maxe;
  char h;

  /*---------------------- OrdFreqTbl ausfllen ----------------------*/
  for (I = 0; I < 256; I++)
    OrdFreqTbl[I] = I;

  for (I = 0; I < 255; I++)
    {
      for (I2 = I + 1, max = -1000, maxe = -1000; I2 < 256; I2++)
	if (ChrFreq[OrdFreqTbl[I2]] > max)
	  {
	    max = ChrFreq[OrdFreqTbl[I2]];
	    maxe = I2;
	  }
      if (maxe == -1000)
	continue;

      if (ChrFreq[OrdFreqTbl[I]] < max)	/* Tauschen ? */
	{
	  h = OrdFreqTbl[maxe];
	  OrdFreqTbl[maxe] = OrdFreqTbl[I];
	  OrdFreqTbl[I] = h;
	}
    }

  /*---------------------- OrdChrTbl ausfllen ----------------------*/

  for (I = 0; I < 256; I++)
    for (I2 = 0; I2 < DifferentChars; I2++)
      if (OrdFreqTbl[I2] == I)
	{
	  OrdChrTbl[I] = I2;
	  break;
	}
}

void
WriteOrdFreqTbl ()
{
  register int I;

  fputc (DifferentChars - 1, OutFile);

  for (I = 0; I < DifferentChars; I++)
    fputc (OrdFreqTbl[I], OutFile);
}


void
ReadOrdChrTbl ()
{
  register int I;
  DifferentChars = fgetc (InFile) + 1;

  for (I = 0; I < DifferentChars; I++)
    OrdChrTbl[I] = fgetc (InFile);
}


/*********************************************************************
*                              Tools  			             *
*********************************************************************/

long
fsize (FILE * f)
{
  long p, len;

  p = ftell (f);
  fseek (f, 0, SEEK_END);
  len = ftell (f);
  fseek (f, p, SEEK_SET);
  return (len);
}



/*********************************************************************
*		            Komprimierung                            *
*********************************************************************/


void
Pack ()
{
  char RLE, Display = 1;

  unsigned RepLen, ThisRepLen, RepPos, ThisRepPos;
  int Opt, ThisOpt;

  static unsigned BitsBuffer[MAXMATCHLENGTH];

  InitRotBuffer ();

  InitWrite ();

  while ((RBuffPos & 32767) != (RBuffEnd & 32767))
    {
      getBitsBuff (BitsBuffer);

      for (Opt = RepLen = 0, ThisRepPos = 1;;)
	{
	  ThisRepLen = RepLen;
	  ThisRepPos = ScanRotRep (&ThisRepLen, ThisRepPos + 1);
	  if (ThisRepPos == 0xFFFF)
	    break;

	  ThisOpt = BitsBuffer[ThisRepLen - 1] -
	    getBitsValue3_1 (ThisRepPos - 1 + DifferentChars) -
	    getBitsValue1_1 (ThisRepLen - 2);

	  if (ThisOpt > Opt)
	    {
	      Opt = ThisOpt;
	      RepPos = ThisRepPos;
	      RepLen = ThisRepLen;
	    }
	}

      ThisRepLen = ScanRotRLE ();
      RLE = 0;
      if (ThisRepLen > 2)
	{
	  ThisOpt = (ThisRepLen - 1) * BitsBuffer[0] -
	    getBitsValue3_1 (DifferentChars) -
	    getBitsValue1_1 (ThisRepLen - 3);
	  if (ThisOpt > Opt)
	    {
	      RLE = 1;
	      RepLen = ThisRepLen;
	    }
	}

      if (!RepLen)		/* keine Wiederholung - Bitkomprimierg. */
	{
	  WriteValue3_1 (OrdChrTbl[RotBuffer[RBuffPos++ & 32767]]);
	  PackedBytes++;
	}
      else if (RLE)
	{
	  WriteValue3_1 (OrdChrTbl[RotBuffer[RBuffPos & 32767]]);
	  WriteValue3_1 (DifferentChars);
	  WriteValue1_1 (RepLen - 3);
	  RBuffPos += RepLen;
	  PackedBytes += RepLen;
	}
      else			/* Wiederholungskomprimierung */
	{
	  WriteValue3_1 (RepPos - 1 + DifferentChars);
	  WriteValue1_1 (RepLen - 2);
	  RBuffPos += RepLen;
	  PackedBytes += RepLen;
	}
      if (!(Display++ & OUTPUTCYCLE))
	{
	  printf
	    ("\r\t Compressing...   %5li >>>> %5li = %6.2f%%;  %5.1f%% completed ",
	     PackedBytes, ftell (OutFile),
	     (double) ((ftell (OutFile) - OutStart) * 100) / PackedBytes,
	     (double) (PackedBytes * 100) / InFileLen);
	  fflush (stdout);
	}

      FillRotBuffer ();
    }
  EndWrite ();

  free (RotBuffer);
}


void
PackOneFile (FILE * In, FILE * Out, unsigned MMDist)
{
  InFile = In;
  OutFile = Out;
  MaxMDist = MMDist;

  InFileLen = fsize (InFile);

  fputc ((InFileLen >> 8) & 0xFF, OutFile);
  fputc (InFileLen & 0xFF, OutFile);

  memset (ChrFreq, 0, 1024);
  DifferentChars = 0;
  OutStart = 0;
  getChrFreq ();
  fseek (InFile, 0, SEEK_SET);
  getOrdChrFreqTbl ();

  WriteOrdFreqTbl ();

  Pack ();
}

void
WriteByte (unsigned char b)
{
  fputc (b, OutFile);
  RotBuffer[RBuffPos++ & 32767] = b;
  UnPackedBytes++;
}

void
UnPack ()
{
  register unsigned I, I2;
  unsigned Value;

  unsigned char c, Display = 0;

  if ((RotBuffer = malloc (32767)) == NULL)
    {
      printf ("\n\n\t ERROR: not enough memory\n\n\t\t program aborted.\n");
      exit (1);
    }
  RBuffPos = 0;
  UnPackedBytes = 0;

  InitRead ();

  while (UnPackedBytes < InFileLen)
    {
      Value = ReadValue3_1 ();
      if (Value < DifferentChars)
	WriteByte (OrdChrTbl[Value]);
      else if (Value == DifferentChars)
	{
	  c = RotBuffer[(RBuffPos - 1) & 32767];
	  for (I = ReadValue1_1 () + 2; I > 0; I--)
	    WriteByte (c);
	}
      else
	{
	  for (I2 = RBuffPos - Value - 1 + DifferentChars, I =
	       ReadValue1_1 () + 2; I > 0; I--, I2++)
	    WriteByte (RotBuffer[I2 & 32767]);
	}
      if (!(Display++ & OUTPUTCYCLE))
	printf ("\r\t Uncompressing... %6.2f%% completed",
		(double) UnPackedBytes * 100 / InFileLen);

    }

  free (RotBuffer);
}

void
UnPackOneFile (FILE * In, FILE * Out)
{
  InFile = In;
  OutFile = Out;

  InFileLen = (long) fgetc (InFile) << 24;
  InFileLen |= (long) fgetc (InFile) << 16;
  InFileLen |= (long) fgetc (InFile) << 8;
  InFileLen |= (long) fgetc (InFile);

  ReadOrdChrTbl ();

  UnPack ();
}

/************************************************************************
*                         Archiverstellung				*
************************************************************************/

FILE *Header, *Temp;

unsigned char *OutName, **InName, HeaderName[256], *TempName;


unsigned InFiles;

enum ARCTYPES
{
  DIVIDED,
  NORMAL,
  SOLID,
  PROGDATA,
};


void
OpenInFile (unsigned n)
{
  if ((InFile = xopen (InName[n], "rb")) == NULL)
    {
      fprintf (stderr, "\n\n\tERROR: Can't open source file \"%s\".",
	       InName[n]);
      exit (1);
    }
}

void
OpenHeader ()
{
  unsigned char *FileExt;
  register int I;

  strcpy (HeaderName, OutName);
  for (I = strlen (HeaderName) - 1, FileExt = &HeaderName[I + 1];
       I > 0 && HeaderName[I] != '\\' && HeaderName[I] != ':'; I--)
    if (HeaderName[I] == '.')
      FileExt = &HeaderName[I];

  strcpy (FileExt, ".inc");

  if ((Header = xopen (HeaderName, "wt")) == NULL)
    {
      fprintf (stderr, "\n\n\tERROR: Can't create include file %s.",
	       HeaderName);
      exit (1);
    }
  fprintf (Header,
	   "\n*************************************************************************");
  fprintf (Header,
	   "\n*            PACKER92 - by David Khling 1997/98 - Header File          *");
  fprintf (Header,
	   "\n*-----------------------------------------------------------------------*");
  fprintf (Header, "\n*  Archive: %-40s                    *", HeaderName);
  fprintf (Header,
	   "\n*************************************************************************\n\n");

  printf ("\n\tCreating Include file %s... completed.", HeaderName);
}

char *
TempNam (char *Dir, char *Pre)	/* For compatability with the TC 3.o source */
{
  char *Nam, *Num, *env;

  unsigned TempNum = 0;

  Nam = malloc (128);
  Nam[0] = '\0';

  if ((env = getenv ("TMP")) == NULL)
    env = getenv ("TEMP");
  if (env != NULL)
    Dir = env;

  if (Dir != NULL)
    {
      strcat (Nam, Dir);
      strcat (Nam, "\\");
    }
  strcat (Nam, Pre);

  Num = &Nam[strlen (Nam)];
  strcpy (Num, "000.$$$");

  while (file_exists_p (Nam))
    {
      TempNum++;
      Num[0] = (TempNum / 100) % 10 + '0';
      Num[1] = (TempNum / 10) % 10 + '0';
      Num[2] = TempNum % 10 + '0';
      if (TempNum > 999)
	{
	  Num[4] = (TempNum / 100000) % 10 + '0';
	  Num[5] = (TempNum / 10000) % 10 + '0';
	  Num[6] = (TempNum / 1000) % 10 + '0';
	}
    }

  return (Nam);
}

void
CreateTemp ()
{
  TempName = TempNam (".", "p92");

  if ((Temp = xopen (TempName, "wb+")) == NULL)
    {
      fprintf (stderr,
	       "\n\n\tERROR: Can't create temporary file.\n\n\t\t\tProgram aborted.\n");
      exit (1);
    }
}

void
RemoveTemp ()
{
  xclose (Temp);
  unlink (TempName);
}

unsigned char *
getFileName (unsigned char *Path, unsigned char *Name)
{
  register int I;
  unsigned char *NamePart;

  for (I = strlen (Path) - 1; I >= 0 &&
       Path[I] != '\\' && Path[I] != ':'; I--);

  NamePart = &Path[++I];

  for (I = 0; I <= strlen (NamePart); I++)
    {
      if (NamePart[I] == '.')
	Name[I] = '_';
      else
	Name[I] = NamePart[I];
    }

  return (Name);
}

/****************************************************************
*                         Solid Archive				*
****************************************************************/

void
PackSolArc ()
{
  register unsigned char c, I;
  static unsigned char ConstName[256];

  CreateTemp ();

  /****************** Include Datei erstellen *******************/

  fprintf (Header, "\n\t;Archive type: SOLID\n");
  for (I = 0; I < InFiles; I++)
    {
      fprintf (Header, "\n%s EQU %ld", getFileName (InName[I], ConstName),
	       ftell (Temp));
      printf ("\n\tReading %s...", InName[I]);

      for (OpenInFile (I);;)
	{
	  c = fgetc (InFile);
	  if (feof (InFile))
	    break;
	  fputc (c, Temp);

	}
      xclose (InFile);

      printf (" completed.");
    }
  fseek (Temp, 0, SEEK_SET);

  /********************** Dateien packen **********************/
  printf ("\n\tCreating archive %s... \n", OutName);
  fputc ('s', OutFile);
  fputc (0, OutFile);
  PackOneFile (Temp, OutFile, 31700);
  printf
    ("\r\t   completed.                                                        ");
  printf
    ("\n\n\t**** %ld >>>> %ld = %6.2f%% optimized. (ratio: %6.2f%%) ****",
     ftell (Temp), ftell (OutFile),
     ((double) ftell (OutFile) * 100) / ftell (Temp),
     ((double) (ftell (Temp) - ftell (OutFile)) * 100) / ftell (Temp));

  RemoveTemp ();
}


/****************************************************************
*                         Normal Archive			*
****************************************************************/

void
PackNormArc ()
{
  register int I;
  static unsigned char ConstName[256];
  long InFilesLen = 0;

  /****************** ChrFreqTbl erstellen *****************/
  /******************** Header erstellen *******************/
  memset (ChrFreq, 0, 1024);
  DifferentChars = 0;

  fprintf (Header, "\n\tArchive type: NORMAL\n");
  for (I = 0, printf ("\n"); I < InFiles; I++)
    {
      printf ("\r\tScanning %-62s\n", InName[I]);
      OpenInFile (I);
      InFileLen = fsize (InFile);
      getChrFreq ();

      fprintf (Header, "\n%s EQU %i", getFileName (InName[I], ConstName), I);
      fprintf (Header, "\n%sl EQU %ld", ConstName, ftell (InFile));
      xclose (InFile);
    }
  printf ("\r\tScanning completed.                                ");

  /********************* Dateien komprimieren ***************/
  fputc ('n', OutFile);
  fputc (InFiles, OutFile);
  for (I = 0; I < InFiles; I++)
    {
      fputc (0, OutFile);
      fputc (0, OutFile);
    }

  getOrdChrFreqTbl ();
  WriteOrdFreqTbl ();

  for (I = 0; I < InFiles; I++)
    {
      printf ("\n\tCompressing %s\n", InName[I]);

	/************ Positionseintrag im Vorspann erstellen **********/
      if (ftell (OutFile) & 1)
	fputc (0, OutFile);

      OutStart = ftell (OutFile);
      fseek (OutFile, (I << 1) + 2, SEEK_SET);
      fputc (OutStart >> 8 & 0xFF, OutFile);
      fputc (OutStart & 0xFF, OutFile);
      fseek (OutFile, OutStart, SEEK_SET);

       /******************* Datei packen ******************************/
      OpenInFile (I);
      InFileLen = fsize (InFile);
      fputc (InFileLen >> 8 & 0xFF, OutFile);
      fputc (InFileLen & 0xFF, OutFile);

      fseek (InFile, 0, SEEK_SET);
      Pack ();

      printf
	("\r\t %ld >>>> %ld = %6.2f%% optimized. (ratio: %6.2f%%)               ",
	 InFileLen, ftell (OutFile) - OutStart,
	 ((double) (ftell (OutFile) - OutStart) * 100) / InFileLen,
	 ((double) (InFileLen - ftell (OutFile) + OutStart) * 100) /
	 InFileLen);
      InFilesLen += InFileLen;
      xclose (InFile);
    }
  printf
    ("\n\n\t**** %ld >>>> %ld = %6.2f%% optimized. (ratio: %6.2f%%) ****",
     InFilesLen, ftell (OutFile),
     ((double) ftell (OutFile) * 100) / InFilesLen,
     ((double) (InFilesLen - ftell (OutFile)) * 100) / InFilesLen);
}

/****************************************************************
*                         Divided Archive			*
****************************************************************/


void
PackDivArc ()
{
  unsigned char ConstName[256];
  register int I;
  long InFilesLen = 0;

  fputc ('d', OutFile);

  /***************** Positionstabelle erzeugen **************/
  fputc (InFiles, OutFile);
  for (I = 0; I < InFiles; I++)
    {
      fputc (0, OutFile);
      fputc (0, OutFile);
    }

  /********************* Dateien packen *********************/
  for (I = 0; I < InFiles; I++)
    {
      OpenInFile (I);
      InFileLen = fsize (InFile);

      fprintf (Header, "\n%s EQU %i", getFileName (InName[I], ConstName), I);
      fprintf (Header, "\n%sl EQU %ld", ConstName, InFileLen);

      memset (ChrFreq, 0, 1024);
      DifferentChars = 0;

      if (ftell (OutFile) & 1)
	fputc (0, OutFile);

      OutStart = ftell (OutFile);
      fseek (OutFile, (I << 1) + 2, SEEK_SET);
      fputc (OutStart >> 8 & 0xFF, OutFile);
      fputc (OutStart & 0xFF, OutFile);
      fseek (OutFile, OutStart, SEEK_SET);

      printf ("\n\tScanning %s\n", InName[I]);
      getChrFreq ();
      fseek (InFile, 0, SEEK_SET);

      fputc ((InFileLen >> 8) & 0xFF, OutFile);
      fputc (InFileLen & 0xFF, OutFile);

      getOrdChrFreqTbl ();
      WriteOrdFreqTbl ();

      printf ("\r\tCompressing %-58s\n", InName[I]);
      Pack ();
      printf
	("\r\t %ld >>>> %ld = %6.2f%% optimized. (ratio: %6.2f%%)               ",
	 InFileLen, ftell (OutFile) - OutStart,
	 ((double) (ftell (OutFile) - OutStart) * 100) / InFileLen,
	 ((double) (InFileLen - ftell (OutFile) + OutStart) * 100) /
	 InFileLen);
      InFilesLen += InFileLen;
      xclose (InFile);
    }

  printf
    ("\n\n\t**** %ld >>>> %ld = %6.2f%% optimized. (ratio: %6.2f%%) ****",
     InFilesLen, ftell (OutFile),
     ((double) ftell (OutFile) * 100) / InFilesLen,
     ((double) (InFilesLen - ftell (OutFile)) * 100) / InFilesLen);
}

/****************************************************************
*                     Program-Data Archive			*
****************************************************************/

unsigned Line;

void
ScanRemark ()
{
  while (fgetc (InFile) != '\r' && !feof (InFile));
  fgetc (InFile);
  Line++;
}

unsigned char *
getPhrase (unsigned char *Phrase)
{
  register int c, I;

  for (;;)
    {
      c = fgetc (InFile);

      if (feof (InFile))
	{
	  Phrase[0] = '\0';
	  return (Phrase);
	}
      if (c == '\r')
	{
	  fgetc (InFile);
	  Line++;
	  continue;
	}
      else if (c == ';')
	{
	  ScanRemark ();
	  continue;
	}
      if (c != ' ' && c != '\t')
	break;
    }


  I = 0;
  do
    {
      if (c == '\r')
	{
	  fgetc (InFile);
	  Line++;
	  break;
	}
      else if (c == ';')
	{
	  ScanRemark ();
	  break;
	}
      Phrase[I++] = c;

      if (I >= 256)
	{
	  fprintf (stderr, "\n\n\t[%i] ERROR: Expression too long:\n%s\n",
		   Line, Phrase);
	  exit (1);
	}
    }
  while ((c = fgetc (InFile)) != ' ' && c != '\t' && !feof (InFile));

  Phrase[I] = '\0';
  return (Phrase);
}

enum DATATYPE
{
  BYTE,
  WORD,
  LONGWORD,
  BYTES,
  WORDS,
  LONGWORDS,
  UNKNOWN
};


void
WriteNum (long Num, enum DATATYPE type)
{
  switch (type)
    {
    case BYTE:
      if (Num < 256 && Num >= -128)
	{
	  fputc (Num, Temp);
	  return;
	}
      break;
    case WORD:
      if (ftell (Temp) & 1)
	printf ("\n\t[%i] WARNING: Word at odd address.", Line);

      if (Num < 65636 && Num >= -32768)
	{
	  fputc (Num >> 8 & 0xFF, Temp);
	  fputc (Num & 0xFF, Temp);
	  return;
	}
      break;
    case LONGWORD:
      if (ftell (Temp) & 1)
	printf ("\n\t[%i] WARNING: Longword at odd address.", Line);

      fputc (Num >> 24 & 0xFF, Temp);
      fputc (Num >> 16 & 0xFF, Temp);
      fputc (Num >> 8 & 0xFF, Temp);
      fputc (Num & 0xFF, Temp);
      return;
    }

  fprintf (stderr, "\n\n\t[%i] ERROR: Numerical overflow.\n", Line);
  exit (1);
}

void
WriteSto (long Num, enum DATATYPE type)
{
  register unsigned int I;

  switch (type)
    {
    case BYTES:
      break;
    case WORDS:
      Num <<= 1;
      if (ftell (Temp) & 1)
	printf ("\n\t[%i] WARNING: Word at odd address.", Line);
      break;
    case LONGWORDS:
      Num <<= 2;
      if (ftell (Temp) & 1)
	printf ("\n\t[%i] WARNING: Longword at odd address.", Line);
    }

  for (I = 0; I < Num; I++)
    fputc (0, Temp);
}

long
ReadDez ()
{
  long Num, Val;

  register int I, c;
  unsigned char NumStr[32];
  unsigned char Neg = 0, NumLen;

  c = fgetc (InFile);
  if (c == '-')
    {
      Neg = 1;
      c = fgetc (InFile);
    }

  for (I = 0;;)
    {
      if (c < '0' || c > '9')
	break;
      if (I >= 32)
	{
	  fprintf (stderr, "\n\n\t[%i] ERROR: Numerical overflow.\n", Line);
	  exit (1);
	}
      NumStr[I++] = c - '0';
      c = fgetc (InFile);
    }
  if (c != EOF)
    ungetc (c, InFile);
  NumLen = I - 1;

  for (I = NumLen, Val = 1, Num = 0; I >= 0; I--, Val *= 10)
    Num += Val * NumStr[I];

  if (Neg)
    Num = -Num;

  return (Num);
}


long
ReadHex ()
{
  long Num, Val;

  register int I, c;
  static unsigned char NumStr[32], Neg = 0, NumLen;

  fgetc (InFile);		/* remove '$'   */
  c = fgetc (InFile);
  if (c == '-')
    {
      Neg = 1;
      c = fgetc (InFile);
    }

  for (I = 0;;)
    {
      if (c < '0' || c > 'F')
	break;
      if (I >= 32)
	{
	  fprintf (stderr, "\n\n\t[%i] ERROR: Numerical overflow.\n", Line);
	  exit (1);
	}
      NumStr[I++] = c - '0';
      c = fgetc (InFile);
    }
  if (c != EOF)
    ungetc (c, InFile);
  NumLen = I - 1;

  for (I = NumLen, Val = 0, Num = 0; I >= 0; I--, Val += 4)
    Num |= NumStr[I] << Val;

  if (Neg)
    Num = -Num;

  return (Num);
}

long
ReadBin ()
{
  long Num, Val;

  register int I, c;
  static unsigned char NumStr[32], Neg = 0, NumLen;

  fgetc (InFile);		/* remove '%'   */

  c = fgetc (InFile);
  if (c == '-')
    {
      Neg = 1;
      c = fgetc (InFile);
    }

  for (I = 0;;)
    {
      if (c < '0' || c > '1')
	break;
      if (I >= 32)
	{
	  fprintf (stderr, "\n\n\t[%i] ERROR: Numerical overflow.\n", Line);
	  exit (1);
	}
      NumStr[I++] = c - '0';
      c = fgetc (InFile);
    }
  if (c != EOF)
    ungetc (c, InFile);
  NumLen = I - 1;

  for (I = NumLen, Val = 0, Num = 0; I >= 0; I--, Val++)
    Num |= NumStr[I] << Val;

  if (Neg)
    Num = -Num;

  return (Num);
}


void
ParseStr (enum DATATYPE type)
{
  unsigned char ec;
  register int c;

  ec = fgetc (InFile);
  if (type != BYTE)
    {
      fprintf (stderr, "\n\n\t[%i] ERROR: String is not allowed here.\n",
	       Line);
      exit (1);
    }

  while ((c = fgetc (InFile)) != ec)
    {
      if (c == '\r' || feof (InFile))
	{
	  fprintf (stderr, "\n\n\t[%i] ERROR: Missing %c\n", Line, ec);
	  exit (1);
	}
      fputc (c, Temp);
    }
}

void
ParseExpr (enum DATATYPE type)
{
  register int c;

  for (;;)
    {

      while ((c = fgetc (InFile)) == ' ' || c == '\t')
	if (feof (InFile) || c == '\r' || c == ';')
	  {
	    fprintf (stderr, "\n\n\t[%i] ERROR: Value expected.\n", Line);
	    exit (1);
	  }

      ungetc (c, InFile);

      if (type <= LONGWORD)
	{
	  if (c == '%')
	    WriteNum (ReadBin (), type);
	  else if (c == '$')
	    WriteNum (ReadHex (), type);
	  else if ((c >= '0' && c <= '9') || c == '-')
	    WriteNum (ReadDez (), type);
	  else if (c == '"' || c == '\'')
	    ParseStr (type);
	  else
	    {
	      fprintf (stderr, "\n\n\t[%i] ERROR: Syntax.\n", Line);
	      exit (1);
	    }
	}
      else if (c == '%')
	WriteSto (ReadBin (), type);
      else if (c == '$')
	WriteSto (ReadHex (), type);
      else if (c >= '0' && c <= '9')
	WriteSto (ReadDez (), type);
      else
	{
	  fprintf (stderr, "\n\n\t[%i] ERROR: Syntax.\n", Line);
	  exit (1);
	}

      do
	{
	  c = fgetc (InFile);

	  if (feof (InFile))
	    return;
	  else if (c == '\r')
	    {
	      Line++;
	      fgetc (InFile);
	      return;
	    }
	  else if (c == ';')
	    {
	      ScanRemark ();
	      return;
	    }
	}
      while (c == ' ' || c == '\t');

      if (c != ',')
	{
	  fprintf (stderr, "\n\n\t[%i] ERROR: Syntax.\n", Line);
	  exit (1);
	}
    }
}

enum DATATYPE
getDataType (unsigned char *Phrase)
{
  if (!strncasecmp (Phrase, "dc.", 3))
    {
      if (!strcasecmp (&Phrase[3], "b"))
	return (BYTE);
      else if (!strcasecmp (&Phrase[3], "w"))
	return (WORD);
      else if (!strcasecmp (&Phrase[3], "l"))
	return (LONGWORD);
    }
  else if (!strncasecmp (Phrase, "ds.", 3))
    if (!strcasecmp (&Phrase[3], "b"))
      return (BYTES);
    else if (!strcasecmp (&Phrase[3], "w"))
      return (WORDS);
    else if (!strcasecmp (&Phrase[3], "l"))
      return (LONGWORDS);

  return (UNKNOWN);
}

unsigned char
LabelOk (unsigned char *l)
{
  if (*l >= '0' && *l <= '9')
    return (0);

  while (*l != '\0')
    {
      if (!
	  ((*l >= '0' && *l <= '9') || (*l >= 'a' && *l <= 'z')
	   || (*l >= 'A' && *l <= 'Z') || *l == '_'))
	return (0);
      l++;
    }
  return (1);

}

void
ParseProgdata ()
{
  static unsigned char Phrase[256];
  enum DATATYPE DataType;

  for (Line = 1; !feof (InFile);)
    {
      getPhrase (Phrase);
      if (!Phrase[0])
	continue;
      if (Phrase[strlen (Phrase) - 1] == ':')
	Phrase[strlen (Phrase) - 1] = '\0';

      if ((DataType = getDataType (Phrase)) == UNKNOWN)
	{
	  if (!LabelOk (Phrase))
	    {
	      fprintf (stderr, "\n\n\t[%i] ERROR: Invalid label name.\n",
		       Line);
	      exit (1);
	    }
	  fprintf (Header, "\n%s EQU %ld", Phrase, ftell (Temp));
	  continue;
	}
      ParseExpr (DataType);
    }

}

void
PackProgdataArc ()
{
  if (InFiles > 1)
    {
      fprintf (stderr, "\n\n\t[%i] ERROR: More than one sourcefile.\n");
      exit (1);
    }

  CreateTemp ();

  OpenInFile (0);
  /*********************** Datei parsen ***********************/

  fprintf (Header, "\n\t;Archive type: PROGRAM DATA\n");

  printf ("\n\tParsing %s...", InName[0]);
  ParseProgdata ();
  printf ("\n\t   completed.");

  /********************** Dateien packen **********************/
  printf ("\n\tCreating archive %s... \n", OutName);
  fputc ('s', OutFile);
  fputc (0, OutFile);
  fseek (Temp, 0, SEEK_SET);
  PackOneFile (Temp, OutFile, 31700);
  printf
    ("\r\t   completed.                                                        ");
  printf
    ("\n\n\t**** %ld >>>> %ld = %6.2f%% optimized. (ratio: %6.2f%%) ****",
     ftell (Temp), ftell (OutFile),
     ((double) ftell (OutFile) * 100) / ftell (Temp),
     ((double) (ftell (Temp) - ftell (OutFile)) * 100) / ftell (Temp));

  xclose (InFile);

  RemoveTemp ();
}

/************************************************************************
*                         Haupt-packfunktion                            *
************************************************************************/

void
PackArc (unsigned char *InNm[], unsigned Ins, char *OutNm,
	 enum ARCTYPES ArcType)
{
  InName = InNm;
  InFiles = Ins;
  OutName = OutNm;

  OutFile = xopen (OutName, "wb");

  OpenHeader ();
  MaxMDist = 31700;

  switch (ArcType)
    {
    case DIVIDED:
      PackDivArc ();
      break;
    case NORMAL:
      PackNormArc ();
      break;
    case SOLID:
      PackSolArc ();
      break;
    case PROGDATA:
      PackProgdataArc ();
      break;
    }

  xclose (OutFile);
  xclose (Header);
}
