/* Level converter */
#include <ctype.h>

#include "include.h"
#include "macro.h"
#include "editor.h"
#include "string92.h"

enum ARCTYPES
{
  DIVIDED,
  NORMAL,
  SOLID,
  PROGDATA,
};
void PackArc (unsigned char *InNm[], unsigned Ins, char *OutNm,
	      enum ARCTYPES ArcType);

#define TILE_BASE (25+3)
#define MAX_SIZE 500000
#define MAX_NAME 512
#define MAX_FILE 255

ub *graph;
ub *test;
ul sg = 0, st = 0;

const ub Signature_d[] = { 'S', 'X', 'W', 'D' };
const ub Signature_n[] = { 'S', 'X', 'N', 'X' };

unsigned long
binfilesize (const char *name)
{
  char name2[MAX_NAME];
  sprintf (name2, "%s.bin", name);
  return filesize (name2);
}

char *
remove_begining_ending_space (char *str)
{
  char *val;
  while (isspace (*str))
    str++;
  val = str;
  while (*str != 0)
    str++;
  str--;
  while (isspace (*str))
    str--;
  str[1] = 0;
  return val;
}

ub
search_link (ub grph, ul size)
{
  ul tab[256];
  ub tst;
  ub *a0;
  ub *a1;
  ul i;

  for (i = 0; i < 256; i++)
    tab[i] = 0;
  a0 = test, a1 = graph;
  for (i = 0; i < size; i++, a1++, a0++)
    if (*(a1) == grph)
      tab[*(a0)]++;

  tst = 0;
  for (i = 1; i < 256; i++)
    if (tab[i] > tab[tst])
      tst = i;

  return tst;
}

int
create_level (LEVEL * L, int NoLevel)
{
  FILE *F1, *F2;
  ul i, j;
  uw k, l;
  ul b;
  ub *a0;
  ub *a1;
  char Name[MAX_NAME];
  ub temp[MAX_NAME];
  char Output[MAX_NAME];
  ul len_data;
  char *InNm[MAX_FILE];
  unsigned int Ins;

  printf ("Create a new level : %s\n", L->Level_name);
  printf ("Gfx1: %s Gfx2: %s\n", L->Graph1_file, L->Graph2_file);
  printf ("Tab Test1: %s Graph1: %s Graph2: %s\n",
	  L->Test_file, L->Gfx_file, L->Plan2);
  printf ("Level: %u x %u Sub-level: %u x %u Plan2: %u x %u\n",
	  L->Mx, L->My, L->Sx, L->Sy, L->Sx2, L->Sy2);
  printf ("Memory requirement: %lu bytes\n",
	  (unsigned long)
	  3840 * 3 + 2 * 10000 + 3200
	  + (2 + L->maxgraph) * L->Sx * L->Sy +
	  L->Mx * L->My / (8 * L->Sx * L->Sy) + L->Sx2 * L->Sy2 +
	  binfilesize ((const char *) L->Graph1_file) +
	  WHEN (strcmp
		((const char *) L->Graph1_file,
		 (const char *) L->Graph2_file), 0,
		binfilesize ((const char *) L->Graph2_file)));
  if (L->maxgraph)
    printf ("Maximun graphs required !\n");

  graph = xalloc (sizeof (ub) * MAX_SIZE * 2);
  test = xalloc (sizeof (ub) * MAX_SIZE);

  F1 = xopen ((const char *) L->Gfx_file, "rb");
  F2 = xopen ((const char *) L->Test_file, "rb");
  if (L->maxgraph)
    sg = fread (graph, sizeof (uw), MAX_SIZE, F1);
  else
    sg = fread (graph, sizeof (ub), MAX_SIZE, F1);
  st = fread (test, sizeof (ub), MAX_SIZE, F2);
  xclose (F2);
  xclose (F1);

  if (sg != st)
    {
      printf ("Error : gfx file and test file have different size !\n");
      return -1;
    }

  if (((L->Mx / L->Sx) * L->Sx != L->Mx)
      || ((L->My / L->Sy) * L->Sy != L->My))
    {
      puts ("Error : Sub-levels have incompatible size with level !");
      return -1;
    }

  if (((long) (L->Sx * L->Sy)) > 32768)
    puts ("Warning : Sub-level's size > 32768, so it could crash !!");

  if (L->Sx2 < 32)
    puts ("Warning : Plane B size X < 32 !");

  /* Creation de la table de liens Graph-Test.
     diminue de 20% la taille du niveau compresse ! */
  printf ("Create link table...");
  F1 = xopen ("link.TMP", "wb");
  if (!L->maxgraph)
    {
      for (i = 0; i < 256; i++)
	temp[i] = search_link (i, L->Mx * L->My);
      fwrite (temp, sizeof (ub), 256, F1);
    }
  xclose (F1);
  printf ("Done !\n");

  /* Rewrite the test matrix using the link table */
  if (!L->maxgraph)
    {
      printf ("ReWriting test...\n");
      a0 = graph;
      a1 = test;
      for (i = 00; i < L->Mx * L->My; i++, a1++, a0++)
	if (*(a1) == temp[*a0])
	  *(a1) = 255;
    }

  /* Curring the 2 matrix (gfx and test) in many small ones */
  printf ("Cutting source files...\n");
  if (L->maxgraph)
    b = sizeof (uw);
  else
    b = sizeof (ub);

  for (j = 0; j < L->My / L->Sy; j++)
    for (i = 0; i < L->Mx / L->Sx; i++)
      {
	sprintf (Name, "l%lux%lu.TMP", i, j);
	F1 = xopen (Name, "wb");
	for (k = 0; k < L->Sy; k++)
	  for (l = 0; l < L->Sx; l++)
	    fwrite (graph +
		    (j * L->Mx * L->Sy + k * L->Mx + i * L->Sx + l) * b, b, 1,
		    F1);
	for (k = 0; k < L->Sy; k++)
	  for (l = 0; l < L->Sx; l++)
	    fwrite (test + j * L->Mx * L->Sy + k * L->Mx + i * L->Sx + l,
		    sizeof (ub), 1, F1);
	xclose (F1);
      }
  xfree (graph);
  xfree (test);

  /* Animation */
  F1 = xopen ("void.TMP", "w");
  xclose (F1);

  /* Creation de la liste de fichier */
  Ins = 0;
  InNm[Ins++] = L->Plan2;
  InNm[Ins++] = "link.TMP";
  InNm[Ins++] = L->Anim_file;
  for (j = 0; j < L->My / L->Sy; j++)
    for (i = 0; i < L->Mx / L->Sx; i++)
      {
	InNm[Ins] = xalloc (10);
	sprintf (InNm[Ins++], "l%lux%lu.TMP", i, j);
      }

  /* Invocation de packer92 */
  PackArc ((unsigned char **) InNm, Ins, "pack1.TMP", NORMAL);

  /* Liberation de la memoire */
  while (Ins > 3)
    xfree (InNm[--Ins]);

  /* Le nom du fichier de sortie */
  sprintf (Output, "LEV%u.TMP", NoLevel);

  // Donnees compresses
  a0 = xalloc (sizeof (ub) * MAX_SIZE);
  F1 = xopen ("pack1.TMP", "rb");
  len_data = fread (a0, sizeof (ub), MAX_SIZE, F1);
  xclose (F1);

  /* Ecrit les donnees de base... */
  F1 = xopen (Output, "wb");

  // Format niveau change...
  //1er fichier graphique + type
  memset (Name, 0, 10);
  for (i = 0; ((Name[i] = *(L->Graph1_file + i))) && (i < 8); i++);
  fwrite (Name, sizeof (ub), 9, F1);
  b = L->Type_P1;
  fwrite (&b, sizeof (ub), 1, F1);

  // 2nd fichier grahique + type
  memset (Name, 0, 10);
  for (i = 0; ((Name[i] = *(L->Graph2_file + i))) && (i < 8); i++);
  fwrite (Name, sizeof (ub), 9, F1);
  b = L->Type_P2;
  fwrite (&b, sizeof (ub), 1, F1);

  // Fichier sprite externe + flag special
  memset (Name, 0, 10);
  for (i = 0; i < 8; i++)
    Name[i] = *(L->Sprite_file + i);
  fwrite (Name, sizeof (ub), 9, F1);
  b = L->maxgraph;
  fwrite (&b, sizeof (ub), 1, F1);

  // Nom du niveau
  memset (Name, 0, 20);
  for (i = 0; ((Name[i] = *(L->Level_name + i)) != 0) && (i < 19); i++);
  fwrite (Name, sizeof (ub), 20, F1);

  // MX1 / MY1 / SX1 / SY1
  i = L->Mx / L->Sx, b = i;
  fwrite (&b, sizeof (ub), 1, F1);
  i = L->My / L->Sy, b = i;
  fwrite (&b, sizeof (ub), 1, F1);
  b = 32;
  for (i = 0, j = 1; i < 32; i++, j *= 2)
    if (j == L->Sx)
      b = i;
  printf ("Sub-level size X = 2^%lu ", b);
  fwrite (&b, sizeof (ub), 1, F1);
  b = L->Sy;
  fwrite (&b, sizeof (ub), 1, F1);

  // Sx2 & Sy2
  b = 32;
  for (i = 0, j = 1; i < 32; i++, j *= 2)
    if (j == L->Sx2)
      b = i;
  printf ("Sub-level plan B size X = 2^%lu\n", b);
  fwrite (&b, sizeof (ub), 1, F1);
  b = L->Sy2;
  fwrite (&b, sizeof (ub), 1, F1);

  // Start position
  i = (L->Startx % L->Sx) * 256 + (L->Startx / L->Sx);
  j = (L->Starty % L->Sy) * 256 + (L->Starty / L->Sy);
  fwrite (&i, sizeof (ub), 2, F1);
  fwrite (&j, sizeof (ub), 2, F1);

  // Donnees compressees
  fwrite (a0, sizeof (ub), len_data, F1);
  xfree (a0);
  xclose (F1);

  return 0;
}

int
read_world (FILE * F, char world_name[100], char next_world[100],
	    int *signature)
{
  LEVEL L;
  char Data[MAX_NAME];
  char temp[MAX_NAME];
  char *Option;
  char *a0;
  ul j, k;
  int NoLevel = 0;

  // Valeur par default
  strcpy (L.Level_name, "User file");
  strcpy (L.Anim_file, "void.TMP");
  strcpy (world_name, "dummy");
  strcpy (next_world, "");

  L.maxgraph = 0;
  L.signature = 0;

  while (!(feof (F)))
    {
      fgets (temp, MAX_NAME, F);
      temp[MAX_NAME] = 0;
      a0 = strchr (temp, '\n');
      if (a0 != NULL)
	*a0 = 0;

      a0 = temp;
      Option = NULL;
      for (j = 0; (j < MAX_NAME) && (*a0 != 0); j++, a0++)
	if (*(a0) == '=')
	  {
	    Option = a0 + 1;
	    break;
	  }

      strncpy (Data, temp, j);
      Data[j] = 0;

      /* 2 arguments */
      if (Option != NULL)
	{
	  Option = remove_begining_ending_space (Option);
	  j = sscanf (Option, "%lu", &k);
	  if (strcmp (Data, "graph1") == 0)
	    strcpy (L.Graph1_file, Option);
	  else if (strcmp (Data, "graph2") == 0)
	    strcpy (L.Graph2_file, Option);
	  else if (strcmp (Data, "sprite") == 0)
	    strcpy (L.Sprite_file, Option);
	  else if (strcmp (Data, "plan2") == 0)
	    strcpy (L.Plan2, Option);
	  else if (strcmp (Data, "animation") == 0)
	    strcpy (L.Anim_file, Option);
	  else if (strcmp (Data, "gfx_tab") == 0)
	    strcpy (&(L.Gfx_file[0]), Option);
	  else if (strcmp (Data, "test_tab") == 0)
	    strcpy (L.Test_file, Option);
	  else if (strcmp (Data, "world_name") == 0)
	    {
	      strcpy (world_name, Option);
	    }
	  else if (strcmp (Data, "level_name") == 0)
	    strcpy (L.Level_name, Option);
	  else if (strcmp (Data, "nextlevel") == 0)
	    {
	      strncpy (next_world, Option, 9);
	      next_world[8] = 0;
	    }
	  /* Si on a pu parser l'option en un nbr */
	  if (j == 1)
	    {
	      if (strcmp (Data, "mx") == 0)
		L.Mx = k;
	      else if (strcmp (Data, "my") == 0)
		L.My = k;
	      else if (strcmp (Data, "sx") == 0)
		L.Sx = k;
	      else if (strcmp (Data, "sy") == 0)
		L.Sy = k;
	      else if (strcmp (Data, "sx2") == 0)
		L.Sx2 = k;
	      else if (strcmp (Data, "sy2") == 0)
		L.Sy2 = k;
	      else if (strcmp (Data, "startx") == 0)
		L.Startx = k;
	      else if (strcmp (Data, "starty") == 0)
		L.Starty = k;
	      else if (strcmp (Data, "P1") == 0)
		L.Type_P1 = k;
	      else if (strcmp (Data, "P2") == 0)
		L.Type_P2 = k;
	      else if (strcmp (Data, "maxgraph") == 0)
		L.maxgraph = k;
	    }
	}
      else			/* un seul argument */
	{
	  if (strcmp (Data, "signaturenext") == 0)
	    *signature = 1;
	  else if (strcmp (Data, "endlevel") == 0)
	    {
	      if (create_level (&L, NoLevel++) != 0)
		{
		  printf ("Error when making level\n");
		  return 0;
		}
	    }
	  else if (strcmp (Data, "endworld") == 0)
	    {
	      return NoLevel;
	    }
	}
    }
  return 0;
}

int
make_world (const char *world_name,
	    int NoLevel, const char *next_world, int signature)
{
  ub *temp, *a0, *a1;
  ul i, j;
  int k;
  char Name[MAX_NAME];
  FILE *F;

  temp = xalloc (sizeof (ub) * MAX_DATA);

  printf ("\nMaking world : %s (%u levels)...\n", world_name, NoLevel);

  a0 = temp;

  // Signature
  if (signature != 0)
    for (i = 0; i < 4; i++)
      *(a0++) = Signature_n[i];
  else
    for (i = 0; i < 4; i++)
      *(a0++) = Signature_d[i];

  // Nombre de niveaux
  *(a0++) = 0;
  *(a0++) = (NoLevel & 255);
  i += 2;

  // Niveau suivant
  for (; i < 10 + 2 + 4; i++)
    *(a0++) = next_world[i - 6];

  // Table des offsets vers les levels
  a1 = a0;			// Offset
  a0 += NoLevel * 2;		// Debut des donnees
  i += NoLevel * 2;

  // Donnees
  for (k = 0; k < NoLevel; k++)
    {
      // On recopie les donnees du niveau
      sprintf (Name, "LEV%d.TMP", k);
      F = xopen (Name, "rb");
      j = fread (a0, sizeof (ub), MAX_DATA, F);
      xclose (F);
      // On remplit la table d'offset vers le niveau
      *(a1++) = LSR ((a0 - temp - 4), 8) & 255;
      *(a1++) = (a0 - temp - 4) & 255;
      // On avance les pointeurs
      j += (i & 1);
      a0 += j;
      i += j;
      printf ("Level %d ; size %lu \n", k, j);
    }
  printf ("World length : %lu\n", i);
  // Creation du monde
  Create92S ((char *) temp, i, (char *) world_name);
  // Finish
  xfree (temp);
  return 0;
}
