/*
 * SMA MKGFX - for Ti-89ti/Ti-89/Ti-92+/V200.
 * Copyright (C) 1999-2006 Patrick Pelissier
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA 
 */

#include "include.h"
#include "macro.h"
#include "string92.h"
#include "shrink.h"

#define TILE_BASE (25+3)
#define MAX_SIZE 65536

ub
create_graph (const char *graph_file)
{
  FILE *F1;
  ub *temp;
  char Name[512];
  char *p;
  ul i, j;
  ul offset = 16 * 4 * TILE_BASE;
  const unsigned char *fileData[1];
  long fileSizes[1];
  unsigned char *compressed_data;

  printf ("Create graph file : %s\n", graph_file);
  temp = xalloc (sizeof (ub) * MAX_SIZE);
  F1 = xopen (graph_file, "rb");
  j = fread (temp, sizeof (ub), MAX_SIZE, F1);
  xclose (F1);

  fileData[0] = temp;
  fileSizes[0] = j;
  if (j < offset)
    abort ();
  memset (temp, 0, offset);
  j = Compress ((unsigned char **) fileData, fileSizes, 1, &compressed_data);

  *(temp + 0) = 'S';
  *(temp + 1) = 'X';
  *(temp + 2) = '1';
  *(temp + 3) = '6';
  memcpy (temp + 4, compressed_data, j);

  strcpy (Name, graph_file);
  p = strrchr (Name, '.');
  if (p)
    *p = 0;
  strcat (Name, ".9xs");
  i = Create92S ((char *) temp, j + 4, Name);
  free (temp);
  free (compressed_data);

  return i;
}

int
main (int argc, const char *argv[])
{
  printf ("SMA WDK: Make Graph Tile files from BIN file.\n"
	  "USAGE: %s gfxfile (without extension).\n", argv[0]);
  if (argc < 2)
    return 0;
  return create_graph (argv[1]);
}
