#include "include.h"
#include "allegro.h"
#include "macro.h"

#include "plane.h"
#include "objplane.h"

/*
 * Initalisation et gestion des plans
 */

void
init_plane (Plane * P)
{
  int i;

  P->mat_file[0] = 0;
  P->skip_byte = 0;
  P->tile_name[0] = 0;
  P->link_file[0] = 0;
  P->tile_no = 0;
  P->height = 0;
  P->length = 0;
  P->sub_height = 0;
  P->sub_length = 0;
  P->type = 2;
  P->xs = 0;
  P->ys = 0;
  P->list_no = 1;
  P->list_no2 = 0;
  P->list_s = 0;
  P->link_no = -1;
  P->Plink = NULL;
  P->link_tab = NULL;

  P->pl_tab = NULL;
  P->pl_flag = NULL;

  strcpy (P->mask, "87654321");
  P->mask[31] = 0;
  P->num_bits = 8;
  P->num_abits = 8;
  P->num_tiles = 256;
  for (i = 0; i < 8; i++)
    P->bit_set[i] = i;
  P->bit_fliph = -1;
  P->bit_flipv = -1;

  P->tile = NULL;
}

void
init_plane2 (Plane * P)
{
  P->xs = 0;
  P->ys = 0;
  P->list_no = 1;
  P->list_no2 = 0;
  P->list_s = 0;
}

void
free_plane (Plane * P)
{
  int j;
  if (P->pl_tab)
    free (P->pl_tab);
  if (P->pl_flag)
    free (P->pl_flag);
  if (P->link_tab)
    free (P->link_tab);
  if (P->tile)
    {
      for (j = 0; j < P->num_tiles; j++)
	free ((P->tile)[j]);
      free (P->tile);
    }
}

void
disp_plane (Plane * P, FILE * F1)
{
  fprintf (F1, "%s\n", P->mat_file);
  fprintf (F1, "%s\n", P->tile_name);
  fprintf (F1, "%s\n", P->link_file);
  fprintf (F1, "%u %u %u %u %u\n", P->skip_byte, P->height, P->length,
	   P->sub_height, P->sub_length);
  fprintf (F1, "%u %u %u %u %u\n", P->xs, P->ys, P->list_no, P->list_no2,
	   P->list_s);
  fprintf (F1, "%d\n", P->link_no);
  fprintf (F1, "%u %u %u\n", P->type, P->taille, P->tile_no);
  fprintf (F1, "%s\n", P->mask);
  fprintf (F1, "%d %d\n", P->bit_fliph, P->bit_flipv);
}

void
do_pause ()
{
  alert ("Pause !", "", "", "Ok", NULL, 0, 0);
}

/*
 * Traduction du masque de gestion
 */
void
translate_mask (Plane * P)
{
  static char alpha[] = "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  int i, j, k;
  int num, numa;
  char *a0;

  // Transforme en forme majustcule
  // Calcul du nombre de bit et du nombre de bits utiles
debut:
  numa = num = strlen (P->mask);
  a0 = P->mask;
  for (i = 0; i < num; i++)
    {
      j = a0[i];
      if (((j < '0') || (j > '9')) && ((j < 'A') || (j > 'Z')))
	{
	  if ((j >= 'a') && (j <= 'z'))
	    a0[i] += 'A' - 'a';
	  else
	    {
	      for (k = i + 1; k < num; k++)
		a0[k - 1] = a0[k];
	      num--;
	      numa--;
	    }
	}
      if (j == '0')
	numa--;
    }
  if (num <= 0)
    {
      strcpy (P->mask, "87654321");
      goto debut;
    }

  // Ok c'est bon !
  P->num_bits = num;
  if (numa > 32)
    numa = 32;
  P->num_abits = numa;
  P->num_tiles = 1 << numa;
  printf ("Bits: %d Tile-bits: %d\n", num, numa);

  // Maintenant on scanne a la recherche des caracteres
  // Pour chaque '1' on recherche sa premiere place
  for (i = 0; i < numa; i++)
    {
      k = alpha[i];
      for (j = 0; ((j < num) && (a0[j] != k)); j++);
      P->bit_set[i] = j;
    }
}



// Gestion bit a bit... Pratique ! ... Mais lent !
void
set_bit (unsigned char *data, unsigned long offset, unsigned long value)
{
  register unsigned long dec, num;

  dec = 7 - (offset & 7);
  num = offset >> 3;

  if (value)
    data[num] |= (1 << dec);
  else if (data[num] & (1 << dec))
    data[num] ^= (1 << dec);
}

unsigned char
get_bit (unsigned char *data, unsigned long offset)
{
  register unsigned long dec, num;

  dec = 7 - (offset & 7);
  num = offset >> 3;

  if (data[num] & (1 << dec))
    return 1;
  return 0;
}


void
save_matrix (Plane * P)
{
  static char *Titre = "Save the matrix of the plane as...";
  FILE *F1;
  ul num, pos;
  ub *a0, *tab;
  ul i, j, k, l, m, n;

  // Demande confirmation du nom.
  if (file_select (Titre, P->mat_file, NULL) == 0)
    return;
  printf ("Trying to save %s...\n", P->mat_file);

  // Ouverture
  if ((F1 = fopen (P->mat_file, "wb+")) == NULL)
    {
      printf ("Error : can't open (rwb) \n");
      alert ("Can't open file !", "", "", "Ok", NULL, 0, 0);
      return;
    }

  // Saute les octets du debut
  if (P->skip_byte != 0)
    fseek (F1, P->skip_byte, SEEK_SET);

  // Alloue memoire
  num = P->length * P->height * P->num_bits;
  if (num & 3)
    num = (num >> 3) + 1;
  else
    num = num >> 3;

  tab = a0 = malloc (num);
  if (a0 == NULL)
    {
      fseek (F1, 0, SEEK_END);
      fclose (F1);
      printf ("Memory full for allocating saving tab !\n");
      alert ("Can't save : memory full !", "", "", "Ok", NULL, 0, 0);
      return;
    }

  // Sauvegarde position fichier
  pos = ftell (F1);

  // Lecture fichier (Pour que les bits non explicites, peut etre modifies
  // par une autre app, y compris plane, soit pris en compte).
  if (fread (a0, sizeof (ub), num, F1) < num)
    // On n'a pas lu assez d'element : on remet tout a 0 !
    memset (a0, 0, num);

  // Correction des bits modifies par l'edition
  for (j = 0, l = 0; j < P->height; j++)
    for (i = 0; i < P->length; i++, l += P->num_bits)
      {
	m = get_tile (P, i, j);
	for (k = 0; k < P->num_abits; k++, m = m >> 1)
	  set_bit (a0, l + P->bit_set[k], (m & 1));
	m = get_flag (P, i, j);
	if (P->bit_flipv > 0)
	  set_bit (a0, l + P->bit_flipv, m & 2);
	if (P->bit_fliph > 0)
	  set_bit (a0, l + P->bit_fliph, m & 1);
      }

  // Sauvegarde
  fseek (F1, pos, SEEK_SET);
  fwrite (a0, sizeof (ub), num, F1);

  // Ferme tout...
  free (a0);
  fseek (F1, 0, SEEK_END);
  fclose (F1);
  printf ("Done\n");
  if (P->Plink)
    {
      Titre = "Save linked plane-file ?";
      save_matrix (P->Plink);
      Titre = "Save the matric file as...";
    }
}





// Ok !
void
open_matrix (Plane * P)
{
  FILE *F1;
  ul num, pos;
  ub *a0;
  ul i, j, k, l, m, n;

  // Demande confirmation
  if (file_select ("Open Matrix file...", P->mat_file, NULL) == 0)
    return;

  printf ("Opening Plane: % s ...\n", P->mat_file);
  if ((F1 = fopen (P->mat_file, "rb")) == NULL)
    {
      printf ("Error : can't open (rb) \n");
      return;
    }

  // Saute les octets en lecture
  if (P->skip_byte != 0)
    fseek (F1, P->skip_byte, SEEK_SET);

  // Alloue memoire
  num = P->length * P->height * P->num_bits;
  if (num & 3)
    num = (num >> 3) + 1;
  else
    num = num >> 3;

  a0 = malloc (num);
  if (a0 == NULL)
    {
      fclose (F1);
      printf ("Memory full for allocating saving tab !\n");
      alert ("Can't save : memory full !", "", "", "Ok", NULL, 0, 0);
      return;
    }

  // Lecture fichier 
  fread (a0, sizeof (ub), num, F1);
  fclose (F1);

  // Traduction
  for (j = 0, l = 0; j < P->height; j++)
    for (i = 0; i < P->length; i++, l += P->num_bits)
      {
	m = 0;
	for (k = 0; k < P->num_abits; k++)
	  if (get_bit (a0, l + P->bit_set[k]))
	    m |= (1 << k);
	set_tile (P, i, j, m);
	m = 0;
	if (P->bit_flipv > 0)
	  if (get_bit (a0, l + P->bit_flipv))
	    m = 2;
	if (P->bit_fliph > 0)
	  if (get_bit (a0, l + P->bit_fliph))
	    m |= 1;
	set_flag (P, i, j, m);
      }

  free (a0);
  printf ("Plane opened\n");
}













// Fonction de rappel de la boite de dialogue
char *
listbox_getter (int index, int *list_size)
{
  static char *strings[] = {
    "Tile 8x8 4c", "Tile 16x16 4c", "Tile 32x32 2c"
  };

  if (index < 0)
    {
      *list_size = 3;
      return NULL;
    }
  else
    return strings[index];
}


// Ouverture d'un plan (Ou nouveau...)
// Si flag = 0, on ouvre la boite de dialogue !
ub
open_plane (Plane * P, uw flag)
{
  ul i, j;
  //ub blt = FALSE; // Reallocation du plan ?
  int ret;
  char buf1[MXS], buf2[MXS], buf3[MXS], buf4[MXS], buf5[MXS], buf6[MXS];
  char buf7[MXS], buf8[MXS], buf9[MXS];
  BITMAP **a0;

  DIALOG the_dialog[] = {
    /* (dialog proc)     (x)             (y)             (w)             (h) (fg)  (bg)  (key)   (flags)         (d1)            (d2)    (dp) */
    {d_shadow_box_proc, 20, 20, 300, 220, 255, 0, 0, 0, 0, 0, NULL},	//0
    {d_text_proc, 30, 35, 180, 18, 255, 0, 0, 0, 0, 0, "Select plane features"},	//1
    {d_text_proc, 30, 75, 65, 18, 255, 0, 0, 0, 0, 0, "TileFile:"},	//2
    {d_edit_proc, 110, 75, 100, 18, 255, 0, 0, 0, sizeof (buf4) - 1, 0, buf4}
    ,				//3
    {d_text_proc, 30, 95, 65, 18, 255, 0, 0, 0, 0, 0, "Map File:"}
    ,				//4
    {d_edit_proc, 110, 95, 100, 18, 255, 0, 0, 0, sizeof (P->mat_file) - 1, 0,
     P->mat_file}
    ,				//5
    {d_text_proc, 30, 115, 50, 18, 255, 0, 0, 0, 0, 0, "Length:"}
    ,				//6
    {d_edit_proc, 85, 115, 40, 18, 255, 0, 0, 0, sizeof (buf1) - 1, 0, buf1}
    ,				//7
    {d_text_proc, 125, 115, 50, 18, 255, 0, 0, 0, 0, 0, "height:"}
    ,				//8
    {d_edit_proc, 180, 115, 40, 18, 255, 0, 0, 0, sizeof (buf2) - 1, 0, buf2}
    ,				//9
    {d_text_proc, 30, 125, 50, 18, 255, 0, 0, 0, 0, 0, "SubLen:"}
    ,				//10
    {d_edit_proc, 85, 125, 40, 18, 255, 0, 0, 0, sizeof (buf5) - 1, 0, buf5}
    ,				//11
    {d_text_proc, 125, 125, 50, 18, 255, 0, 0, 0, 0, 0, "SubHht:"}
    ,				//12
    {d_edit_proc, 180, 125, 40, 18, 255, 0, 0, 0, sizeof (buf6) - 1, 0, buf6}
    ,				//13
    {d_list_proc, 25, 145, 150, 40, 255, 0, 0, 0, 1, 0, listbox_getter}
    ,				//14
    {d_button_proc, 30, 210, 60, 18, 255, 0, 0, D_EXIT, 0, 0, "OK"}
    ,				//15
    {d_button_proc, 100, 210, 60, 18, 255, 0, 0, D_EXIT, 0, 0, "Cancel"}
    ,				//16
    {d_text_proc, 30, 190, 50, 18, 255, 0, 0, 0, 0, 0, "Skip bytes:"}
    ,				//17
    {d_edit_proc, 130, 190, 40, 18, 255, 0, 0, 0, sizeof (buf3) - 1, 0, buf3}
    ,				//18
    {d_text_proc, 220, 75, 50, 18, 255, 0, 0, 0, 0, 0, "Mask:"}
    ,				//19
    {d_edit_proc, 220, 85, 100, 18, 255, 0, 0, 0, sizeof (buf7) - 1, 0, buf7}
    ,				//20
    {d_text_proc, 220, 115, 50, 18, 255, 0, 0, 0, 0, 0, "Bit HFlip:"}
    ,				//21
    {d_edit_proc, 220, 125, 40, 18, 255, 0, 0, 0, sizeof (buf8) - 1, 0, buf8}
    ,				//24
    {d_text_proc, 220, 135, 50, 18, 255, 0, 0, 0, 0, 0, "Bit VFlip:"}
    ,				//25
    {d_edit_proc, 220, 145, 40, 18, 255, 0, 0, 0, sizeof (buf9) - 1, 0, buf9}
    ,				//26
    {NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL}
  };

  // Si on demande la boite de dialogue
  if (flag == 0)
    {
      sprintf (buf1, "%u", P->length);
      sprintf (buf2, "%u", P->height);
      sprintf (buf3, "%u", P->skip_byte);
      sprintf (buf4, "%s", P->tile_name);
      sprintf (buf5, "%u", P->sub_length);
      sprintf (buf6, "%u", P->sub_height);
      sprintf (buf7, "%s", P->mask);
      sprintf (buf8, "%d", P->bit_fliph);
      sprintf (buf9, "%d", P->bit_flipv);
      centre_dialog (the_dialog);
      ret = popup_dialog (the_dialog, 1);
      if ((ret == 16) || (ret <= 0))
	return FALSE;

      free_plane (P);

      sscanf (buf1, "%u", &i);
      sscanf (buf2, "%u", &j);
      P->length = i;
      P->height = j;
      sscanf (buf3, "%u", &i);
      P->skip_byte = i;
      sscanf (buf5, "%u", &i);
      P->sub_length = i;
      sscanf (buf6, "%u", &i);
      P->sub_height = i;
      sscanf (buf8, "%d", &i);
      P->bit_fliph = i;
      sscanf (buf9, "%d", &i);
      P->bit_flipv = i;
      switch (the_dialog[14].d1)
	{
	case 0:
	  P->type = 1;
	  P->taille = 8;
	  break;
	case 1:
	  P->type = 2;
	  P->taille = 16;
	  break;
	case 2:
	  P->type = 3;
	  P->taille = 32;
	  break;
	}
      strcpy (P->tile_name, buf4);
      strcpy (P->mask, buf7);
    }
  // Fin de traitement de la boite de dialogue. 

  translate_mask (P);

  // Allocation memoire de la matrice et des flags
  if (P->pl_tab)
    free (P->pl_tab);
  if (P->pl_flag)
    free (P->pl_flag);
  if ((P->pl_tab =
       (ul *) malloc (P->length * P->height * sizeof (ul))) == NULL)
    {
      printf ("Not enought memory to allocate the matrix table\n");
      alert ("Memory full !", "", "", "Ok", NULL, 0, 0);
      return FALSE;
    };
  if ((P->pl_flag =
       (ub *) malloc (P->length * P->height * sizeof (ub))) == NULL)
    {
      printf ("Not enought memory to allocate the flag table\n");
      alert ("Memory full !", "", "", "Ok", NULL, 0, 0);
      return FALSE;
    };

  for (i = 0; i < (P->length * P->height); i++)
    P->pl_tab[i] = 0, P->pl_flag[i] = 0;

  // Ouverture si elle existe
  if (exists (P->mat_file))
    open_matrix (P);

  // Ouverture des tiles
  P->tile = (BITMAP **) malloc (sizeof (BITMAP *) * P->num_tiles);
  if (P->tile == NULL)
    {
      printf ("Not enought memory to allocate the tiles indirect table\n");
      alert ("Memory full !", "", "", "Ok", NULL, 0, 0);
      return FALSE;
    }
  a0 = P->tile;
  for (i = 0; i < P->num_tiles; i++)
    *(a0++) = NULL;

  // Ouverture si ca existe
  if (P->tile_name[0] != 0)
    if ((P->tile_no =
	 load_tile (P->tile_name, P->type, P->taille, P->num_tiles,
		    P->tile)) == 0)
      {
	printf ("Error when loading and creating tiles\n");
	alert ("Can't load tiles", "", "", "ok", NULL, 0, 0);
	return FALSE;
      }

  printf ("Plane opened !\n");
  return TRUE;
}


void
print_Ptile (Plane * P, ul x, ul y, ul no, uw flag)
{
  if (no < P->tile_no)
    print_tile (x, y, no, flag, P->tile);
}




// Ok !
void
save_plane_as_bitmap (Plane * P)
{
  BITMAP *bmp;
  ul i, j, k = 0, pas = P->taille;
  ul num;
  ul flag;


  bmp = create_bitmap (P->length * pas, P->height * pas);	// make a bitmap in system RAM
  for (j = 0; (j < P->height); j++)
    for (i = 0; (i < P->length); i++, k++)
      {
	flag = 0;
	num = get_tile (P, i, j);
	if (num < P->num_tiles)
	  switch (get_flag (P, i, j))
	    {
	    case 3:
	      draw_sprite_vh_flip (bmp, (P->tile)[num], i * pas, j * pas);
	      break;
	    case 2:
	      draw_sprite_v_flip (bmp, (P->tile)[num], i * pas, j * pas);
	      break;
	    case 1:
	      draw_sprite_h_flip (bmp, (P->tile)[num], i * pas, j * pas);
	      break;
	    case 0:
	      draw_sprite (bmp, (P->tile)[num], i * pas, j * pas);
	      break;
	    }
      }
  save_pcx ("bitmap.pcx", bmp, desktop_pallete);
  destroy_bitmap (bmp);
}






// Lecture des tiles
ul
get_tile (Plane * P, ul x, ul y)
{
  if ((x > P->length) || (y > P->height))
    return 0;
  return P->pl_tab[y * P->length + x];
}



// Ecriture des tiles
void
set_tile (Plane * P, ul x, ul y, ul num)
{
  if ((x > P->length) || (y > P->height))
    return;

  P->pl_tab[y * P->length + x] = num;
}

// Lecture du flag
ub
get_flag (Plane * P, ul x, ul y)
{
  if ((x > P->length) || (y > P->height))
    return 0;
  return P->pl_flag[y * P->length + x];
}

// Ecriture du flag
void
set_flag (Plane * P, ul x, ul y, ub num)
{
  if ((x > P->length) || (y > P->height))
    return;

  P->pl_flag[y * P->length + x] = num;
}


void
fliph (Plane * P, ul x, ul y)
{
  set_flag (P, x, y, (get_flag (P, x, y) ^ 1));
}

void
flipv (Plane * P, ul x, ul y)
{
  set_flag (P, x, y, (get_flag (P, x, y) ^ 2));
}







// ok !
void
open_link (Plane * P)
{
  FILE *F1;
  int j;

  free (P->link_tab);

  if ((P->link_tab = malloc (sizeof (ul) * P->num_tiles)) == NULL)
    {
      printf ("Memory : link table ! \n");
      alert ("Error : memory", "Not enought free memory to allocate",
	     "a link table !", "Ok", NULL, 0, 0);
      return;
    }

  for (j = 0; j < P->num_tiles; j++)
    P->link_tab[j] = 0;

  if (file_select ("Open link-file...", P->link_file, NULL) == 0)
    return;
  printf ("Opening link: % s ...\n", P->link_file);

  if ((F1 = fopen (P->link_file, "rb")) == NULL)
    {
      printf ("Error : can't open %s (r) \n", P->link_file);
      return;
    }
  if (fread (P->link_tab, sizeof (long), P->num_tiles, F1) != P->num_tiles)
    {
      printf ("Error :link table too small !\n");
      alert ("Illegal link table !", "", "", "ok", NULL, 0, 0);
    }
  else
    printf ("Link %s opened\n", P->link_file);

  fclose (F1);
}



// Ok !
void
save_link (Plane * P)
{
  FILE *F1;
  if (file_select ("Save link-file...", P->link_file, NULL) == 0)
    return;
  printf ("Saving link: % s ...\n", P->link_file);
  if ((F1 = fopen (P->link_file, "wb")) == NULL)
    {
      printf ("Error : can't open %s (w) \n", P->link_file);
      return;
    }
  fwrite (P->link_tab, sizeof (ul), P->num_tiles, F1);
  fclose (F1);
  printf ("Link %s saved\n", P->link_file);
}


ul
select_link (Plane * P, ul x)
{
  ul y;
  ul i, j, k, t, a;
  ub cont = TRUE;
  if (P->Plink == NULL)
    {
      printf ("Error : no link !\n");
      return x;
    }
  i = k = 0;
  t = P->Plink->taille;
  while (cont)
    {
      y = 0;
      j = i;
      clear (buffer);
      while ((y + t) < SCREEN_H)
	{
	  print_Ptile (P->Plink, 50, y, j, 0);
	  if (j == k)
	    print_selected_tile (50, y, P->taille);
	  j++;
	  y += t + 1;
	}
      scare_mouse ();
      blit (buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
      unscare_mouse ();
      readkey ();
      if (key[KEY_ESC])
	cont = FALSE;
      if (key[KEY_UP])
	if (k > 0)
	  {
	    k--;
	    if (i > k)
	      i = k;
	  }
      if (key[KEY_DOWN])
	if (k < P->Plink->tile_no)
	  {
	    k++;
	    if (k >= j)
	      i++;
	  }
      if (key[KEY_PGUP])
	for (a = 0; a < 15; a++)
	  if (k > 0)
	    {
	      k--;
	      if (i > k)
		i = k;
	    }
      if (key[KEY_PGDN])
	for (a = 0; a < 15; a++)
	  if (k < P->Plink->tile_no)
	    {
	      k++;
	      if (k >= j)
		i++;
	    }
      if (key[KEY_ENTER])
	{
	  x = k;
	  cont = FALSE;
	}
    }
  return x;
}



void
edit_link (Plane * P)
{
  ul x, y;
  ul i, j, k, t, a;
  ub cont = TRUE;
  printf ("Edit link...\n");
  if (P->Plink == NULL)
    {
      printf ("Error : no link !\n");
      return;
    }
  t = MAX (P->taille, P->Plink->taille);
  printf ("MAX taille of link : %u \n", t);
  i = k = 0;
  while (cont)
    {
      y = 0;
      j = i;
      clear (buffer);
      while ((y + t) < SCREEN_H)
	{
	  print_Ptile (P, 10, y, j, 0);
	  if (j == k)
	    print_selected_tile (10, y, P->taille);
	  print_Ptile (P->Plink, 100, y, *(P->link_tab + j), 0);
	  j++;
	  y += t + 1;
	}
      scare_mouse ();
      blit (buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
      unscare_mouse ();
      readkey ();
      if (key[KEY_ESC])
	cont = FALSE;
      if (key[KEY_UP])
	{
	  if (k > 0)
	    {
	      k--;
	      if (i > k)
		i = k;
	    }
	}
      if (key[KEY_DOWN])
	{
	  if (k < P->tile_no)
	    {
	      k++;
	      if (k >= j)
		i++;
	    }
	}
      if (key[KEY_PGUP])
	for (a = 0; a < 15; a++)
	  if (k > 0)
	    {
	      k--;
	      if (i > k)
		i = k;
	    }
      if (key[KEY_PGDN])
	for (a = 0; a < 15; a++)
	  if (k < P->tile_no)
	    {
	      k++;
	      if (k >= j)
		i++;
	    }
      if (key[KEY_ENTER])
	*(P->link_tab + k) = select_link (P, *(P->link_tab + k));
    }
  i = alert ("Save new link file ?", "", "", "Yes", "No", 0, 0);
  if (i == 1)
    save_link (P);
  return;
}
