/*
; PACKER92 - Copyright 1998 David Kuehling
; Adaptation for TI-89/92+ - Copyright 2002, 2003, 2004, 2005 Patrick Pelissier
;
; This file is part of the PACKER92 Library.
;
; The PACKER92 Library is free software; you can redistribute it and/or modify
; it under the terms of the GNU Lesser General Public License as published by
; the Free Software Foundation; either version 2.1 of the License, or (at your
; option) any later version.
;
; The PACKER92 Library is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
; License for more details.
;
; You should have received a copy of the GNU Lesser General Public License
; along with the PACKER92 Library; see the file COPYING.LIB.  If not, write to
; the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
; MA 02111-1307, USA.
*/

#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <io.h>
#include <string.h>

#define ANSI_setAtt(v,h,a)	printf ("\33[%d;%d;%dm", v+30, h+40, a);

enum ARCTYPES
{
  DIVIDED,
  NORMAL,
  SOLID,
  PROGDATA,
};


void
PrintHelp ()
{
  printf ("\n\nHelp:");
  printf
    ("\n\tSyntax: packer92 -X [@]source1 [[@]source2] [...] destination");
  printf ("\n\t -X:");
  printf ("\n\t\t -n  create normal archive");
  printf ("\n\t\t -d  create divided archive");
  printf ("\n\t\t -s  create solid archive");
  printf ("\n\t\t -p  create program data archive");
  printf
    ("\n\t sourceX - name of source file \n\t\t(if -X is not -p, you can include up to 255 sourcefiles)");
  printf
    ("\n\t @sourceX - name of list-file, containing source files, divided by \n\t\tlinebreaks");
  printf
    ("\n\t destination - name of destination file \n\t\t(The file, which is included by INCBIN)");
  printf
    ("\n\t\tA destination file with the extention .inc\n\t\twill be created too, so don't use a .inc file \n\t\tas destination.");
}

void
main (unsigned argn, char *args[])
{
  unsigned I;
  register int c, I2;
  enum ARCTYPES ArcType;
  static unsigned char FileName[256][128], *PtrFlNm[256], FileNames;
  FILE *ListFile;


  printf ("\n\t\t---------------------------------------------");
  printf ("\n\t\t  PACKER92 v1.10 - by David Khling 1997/98  ");
  printf ("\n\t\t---------------------------------------------\n");

  if (argn < 4 || (args[1][0] != '-' && args[1][0] != '/'))
    {
      PrintHelp ();
      exit (1);
    }

  switch (args[1][1])
    {
    case 'n':
      ArcType = NORMAL;
      break;
    case 'd':
      ArcType = DIVIDED;
      break;
    case 's':
      ArcType = SOLID;
      break;
    case 'p':
      ArcType = PROGDATA;
      break;
    default:
      fprintf (stderr, "\n\n\tERROR: Unknown command %s.", args[1]);
      PrintHelp ();
      exit (1);
    }

  /* Copy filenames to FileName */
  for (I = 2, FileNames = 0, PtrFlNm[0] = FileName[0]; I < argn - 1; I++)
    {
      if (args[I][0] != '@')
	{
	  PtrFlNm[FileNames] = FileName[FileNames];
	  strcpy (FileName[FileNames++], args[I]);
	}
      else
	{
	  if ((ListFile = xopen (&args[I][1], "rt")) == NULL)
	    {
	      fprintf (stderr, "\n\n\tERROR: can't open list-file: %s\n",
		       &args[I][1]);
	      exit (1);
	    }
	  for (I2 = 0;;)
	    {
	      if ((c = fgetc (ListFile)) == EOF)
		{
		  if (I2)
		    FileName[FileNames++][I2] = '\0';
		  break;
		}
	      if (c == '\n')
		{
		  if (I2)
		    FileName[FileNames++][I2] = '\0';
		  if (FileNames >= 255)
		    {
		      fprintf (stderr,
			       "\n\n\tERROR: More than 255 sourcefiles.\n");
		      exit (1);
		    }
		  I2 = 0;
		  PtrFlNm[FileNames] = FileName[FileNames];
		  continue;
		}
	      if (I2 >= 127)
		{
		  fprintf (stderr,
			   "\n\n\tERROR: Too long filename in list-file %s.\n",
			   &args[I][1]);
		  exit (1);
		}
	      FileName[FileNames][I2++] = c;
	    }
	  xclose (ListFile);
	}
    }
  PackArc (PtrFlNm, FileNames, args[argn - 1], ArcType);
}
